# Interface to the Salesforce BULK API
import os
from collections import namedtuple
from httplib2 import Http
import requests
import urllib2
import urlparse
import requests
import xml.etree.ElementTree as ET
from tempfile import TemporaryFile, NamedTemporaryFile
import StringIO
import re
import time
import csv

UploadResult = namedtuple('UploadResult', 'id success created error')


def dump_results(data, failed, count):
    if hasattr(data, 'read'):
        print data.read()
    else:
        print data

class BulkApiError(Exception):
    pass

class SalesforceBulk(object):
    def __init__(self, sessionId = None, host=None, username=None, password=None):
        if not sessionId and not username:
            raise RuntimeError("Must supply either sessionId/instance_url or username/password")
        if not sessionId:
            sessionId, endpoint = SalesforceBulk.login_to_salesforce(username, password)
            host = urlparse.urlparse(self.h._location)
            host = host.hostname.replace("-api","")

        if host[0:4] == 'http':
            self.endpoint = host
        else:
            self.endpoint = "https://" + host
        self.sessionId = sessionId
        #print "Creating bulk adapter, endpoint %s, id: %s" % (endpoint, sessionId)
        self.jobNS = 'http://www.force.com/2009/06/asyncapi/dataload'
        self.jobs = {} # dict of job_id => job_id
        self.batches = {} # dict of batch_id => job_id
        self.batch_statuses = {}

    @staticmethod
    def login_to_salesforce(username, password):
        for key in ['SALESFORCE_CLIENT_ID','SALESFORCE_CLIENT_SECRET','SALESFORCE_REDIRECT_URI']:
            if key not in os.environ:
                raise RuntimeError("You must set %s to use username/pass login" % key)

        try:
            import salesforce_oauth_request
        except ImportError:
            raise ImportError("You must install salesforce-oauth-request to use username/password")

        packet = salesforce_oauth_request.login(username=username, password=password)
        return packet['access_token'],packet['instance_url']

    def headers(self, values = {}):
        default = {"X-SFDC-Session" : self.sessionId, "Content-Type" : "application/xml; charset=UTF-8"}
        for k, val in values.iteritems():
            default[k] = val
        return default

    # Register a new Bulk API job - returns the job id
    def create_query_job(self, object_name, **kwargs):
        return self.create_job(object_name, "query", **kwargs)

    def create_insert_job(self, object_name, **kwargs):
        return self.create_job(object_name, "insert", **kwargs)

    def create_update_job(self, object_name, **kwargs):
        return self.create_job(object_name, "update", **kwargs)

    def create_delete_job(self, object_name, **kwargs):
        return self.create_job(object_name, "delete", **kwargs)

    def create_job(self, object_name = None, operation = None, contentType = 'CSV', concurrency=None):
        assert(object_name is not None)
        assert(operation is not None)

        doc = self.create_job_doc(object_name=object_name,
                                  operation=operation,
                                  contentType=contentType,
                                  concurrency=concurrency)

        http = Http()
        resp, content = http.request(self.endpoint + "/services/async/29.0/job", "POST", headers=self.headers(), body=doc)

        self.check_status(resp, content)

        tree = ET.fromstring(content)
        job_id = tree.findtext("{%s}id" % self.jobNS)
        self.jobs[job_id] = job_id

        return job_id

    def check_status(self, resp, content):
        if resp.status >= 400:
            print "Non-200 status: %d" % resp.status
            raise Exception(content)

    def close_job(self, job_id):
        doc = self.create_close_job_doc()
        http = Http()
        resp, content = http.request(self.endpoint + "/services/async/29.0/job/%s" % job_id, "POST",
            headers=self.headers(), body=doc)
        self.check_status(resp, content)


    def create_job_doc(self, object_name = None, operation = None, contentType = 'CSV', concurrency=None):
        root = ET.Element("jobInfo")
        root.set("xmlns", self.jobNS)
        op = ET.SubElement(root, "operation")
        op.text  = operation
        obj = ET.SubElement(root, "object")
        obj.text = object_name
        if concurrency:
            con = ET.SubElement(root, "concurrencyMode")
            con.text = concurrency
        ct = ET.SubElement(root, "contentType")
        ct.text = contentType

        buf = StringIO.StringIO()
        tree = ET.ElementTree(root)
        tree.write(buf, encoding ="UTF-8")
        return buf.getvalue()

    def create_close_job_doc(self, object_name = None, operation = None, contentType = 'CSV'):
        root = ET.Element("jobInfo")
        root.set("xmlns", self.jobNS)
        state = ET.SubElement(root, "state")
        state.text  = "Closed"

        buf = StringIO.StringIO()
        tree = ET.ElementTree(root)
        tree.write(buf, encoding ="UTF-8")
        return buf.getvalue()

    # Add a BulkQuery to the job - returns the batch id
    def query(self, job_id, soql):
        if job_id is None:
            job_id = self.create_job(re.search(re.compile("from (\w+)",re.I), soql).group(1), "query")
        http = Http()
        uri = self.endpoint + "/services/async/29.0/job/%s/batch" % job_id
        resp, content = http.request(uri, method="POST", body=soql, headers=self.headers({"Content-Type":"text/csv"}))

        self.check_status(resp, content)

        tree = ET.fromstring(content)
        batch_id = tree.findtext("{%s}id" % self.jobNS)

        self.batches[batch_id] = job_id

        return batch_id

    def split_csv(self, csv, batch_size):
        csv_io = StringIO.StringIO(csv)
        batches = []

        for i, line in enumerate(csv_io):
            if not i:
                headers = line
                batch = headers
                continue
            if not i % 100:
                print i / 100
            if not i % batch_size:
                print 'New batch at %s' % i
                batches.append(batch)
                batch = headers

            batch += line

        batches.append(batch)

        return batches

    # Add a BulkUpload to the job - returns the batch id
    def bulk_csv_upload(self, job_id, csv, batch_size=2500):
        # Split a large CSV into manageable batches
        batches = self.split_csv(csv, batch_size)
        batch_ids = []

        uri = self.endpoint + "/services/async/29.0/job/%s/batch" % job_id
        for batch in batches:
            print len(batch)
            resp = requests.post(uri, data=batch, headers=self.headers({"Content-Type":"text/csv"}))
            content = resp.content

            if resp.status_code >= 400:
                print "Non-200 status: %d" % resp.status
                raise Exception(content)

            tree = ET.fromstring(content)
            batch_id = tree.findtext("{%s}id" % self.jobNS)

            self.batches[batch_id] = job_id
            batch_ids.append(batch_id)

        return batch_ids

    def post_bulk_batch(self, job_id, csv_generator):
        uri = self.endpoint + "/services/async/29.0/job/%s/batch" % job_id
        resp = requests.post(uri, data=csv_generator, headers=self.headers({"Content-Type":"text/csv"}))
        content = resp.content

        if resp.status_code >= 400:
            print "Non-200 status: %d" % resp.status_code
            raise Exception(content)

        tree = ET.fromstring(content)
        batch_id = tree.findtext("{%s}id" % self.jobNS)
        return batch_id

    # Add a BulkDelete to the job - returns the batch id
    def bulk_delete(self, job_id, object_type, where, batch_size=2500):
        query_job_id = self.create_query_job(object_type)
        soql = "Select Id from %s where %s Limit 10000" % (object_type, where)
        print soql
        query_batch_id = self.query(query_job_id, soql)
        self.wait_for_batch(query_job_id, query_batch_id, timeout=120)

        results = []
        def save_results(tf, **kwargs):
            results.append(tf.read())

        flag = self.get_batch_results(query_job_id, query_batch_id, callback = save_results)

        if job_id is None:
            job_id = self.create_job(object_type, "delete")
        http = Http()
        # Split a large CSV into manageable batches
        batches = self.split_csv(csv, batch_size)
        batch_ids = []

        uri = self.endpoint + "/services/async/29.0/job/%s/batch" % job_id
        for batch in results:
            print len(batch)
            print batch
            print batch.split('\n')
            resp = requests.post(uri, data=batch, headers=self.headers({"Content-Type":"text/csv"}))
            content = resp.content

            if resp.status_code >= 400:
                print "Non-200 status: %d" % resp.status
                raise Exception(content)

            tree = ET.fromstring(content)
            batch_id = tree.findtext("{%s}id" % self.jobNS)

            self.batches[batch_id] = job_id
            batch_ids.append(batch_id)

        return batch_ids

    def lookup_job_id(self, batch_id):
        try:
            return self.batches[batch_id]
        except KeyError:
            raise Exception("Batch id '%s' is uknown, can't retrieve job_id" % batch_id)

    def batch_status(self, job_id = None, batch_id = None, reload=False):
        if not reload and batch_id in self.batch_statuses:
            return self.batch_statuses[batch_id]

        job_id = job_id or self.lookup_job_id(batch_id)

        http = Http()
        uri = self.endpoint + "/services/async/29.0/job/%s/batch/%s" % (job_id, batch_id)
        resp, content = http.request(uri, headers=self.headers())
        self.check_status(resp, content)

        tree = ET.fromstring(content)
        result = {}
        for child in tree:
            result[re.sub("{.*?}","",child.tag)] = child.text

        self.batch_statuses[batch_id] = result
        return result

    def batch_state(self, job_id, batch_id, reload=False):
        status = self.batch_status(job_id, batch_id, reload=reload)
        if 'state' in status:
            return status['state']
        else:
            return None

    def is_batch_done(self, job_id, batch_id):
        state = self.batch_state(job_id, batch_id, reload=True)
        print "BULK STATE IS: %s" % state
        if state == 'Failed' or state == 'Not Processed':
            status = self.batch_status(job_id, batch_id)
            if state == 'Failed':
                self.close_job(job_id)
            raise BulkApiError("Batch %s of job %s failed: %s" % (batch_id, job_id, status['stateMessage']))
        return state == 'Completed'

    # Wait for the given batch to complete, waiting at most timeout seconds (defaults to 10 minutes).
    def wait_for_batch(self, job_id, batch_id, timeout = 60*10, sleep_interval=10):
        waited = 0
        while not self.is_batch_done(job_id, batch_id) and waited < timeout:
            time.sleep(sleep_interval)
            waited += sleep_interval


    # Retrieve the results of a batch query. Your callback will be invoked 0 or more times
    # with a TemporaryFile object from which you can read the returned data. If the batch is not
    # ready then your callback may not be invoked.
    # Returns False if the batch isn't Completed, or True if it is and was processed.
    #
    # By default your callback will receive a File-like object with the raw results. You can
    # pass 'parse_csv = True' to request that CSV data be parsed. In this case your callback
    # will receive two parameters. The first will be a list of successfully parsed rows.
    # Each row will contained a list of values. The column names will appear as the first row.
    # So for csv data:
    #
    #   "field1","field2"
    #   "row1col1","row1col2"
    #
    # You will receive: [["field1","field2"],["row1col1","row2col2"]]
    #
    # The second argument will be a list of records which failed parsing. The format of each row is:
    #    [<line number>,<error message>, <csv record>]

    # You can pass batch_size=N to have your callback invoked with batches of at most N rows. Note
    # that row[0] of *every* batch will hold the list of columns names.
    #
    # This call is resumable. You should record the 'line' (int) and 'result_id' (str) passed to your callback,
    # and if you need to restart in the middle you can pass those values back as 'skip_lines' and
    # 'skip_to_result' in the call. We will move ahead to the indicated result file, then skip
    # the count of lines before sending additional rows to your callback.

    def get_batch_results(self, job_id, batch_id, callback = dump_results, parse_csv=False, batch_size=0,
                          skip_lines=None, skip_to_result=None, logger=None):
        job_id = job_id or self.lookup_job_id(batch_id)

        if not self.is_batch_done(job_id, batch_id):
            return False
        http = Http()
        uri = self.endpoint + "/services/async/29.0/job/%s/batch/%s/result" % (job_id, batch_id)
        resp, content = http.request(uri, method="GET", headers=self.headers())

        tree = ET.fromstring(content)
        for result in tree.iterfind("{%s}result" % self.jobNS):
            result_id = str(result.text)

            if skip_to_result and result_id != skip_to_result:
                logger("Skipping processed bulk result '%s'" % result_id)
                continue

            ruri = self.endpoint + "/services/async/29.0/job/%s/batch/%s/result/%s" % (job_id, batch_id, result_id)
            if logger:
                logger("Downloading bulk result file: %s" % result_id)
                #print "$$$$$$$$$$ Downloading bulk result file: %s" % ruri
            resp = requests.get(ruri, headers=self.headers(), stream=True)
            tf = TemporaryFile()
            for chunk in resp.iter_content(1024):
                tf.write(chunk) # assume we're writing utf8
                print "##"
            resp.close()

            total_remaining = self.count_file_lines(tf)
            if skip_lines:
                total_remaining -= skip_lines
                logger("Skipping %d lines of bulk file" % skip_lines)

            if logger:
                logger("Total records: %d" % total_remaining)

            tf.seek(0)
            if not parse_csv:
                callback(tf)
            else:
                records = []
                line_number = 0
                col_names = []
                reader = csv.reader(tf, delimiter=",", quotechar='"')
                for row in reader:
                    line_number += 1
                    if line_number > 1 and skip_lines and line_number <= skip_lines:
                        continue
                    records.append(row)
                    if len(records) == 1:
                        col_names = records[0]
                    if batch_size > 0 and len(records) >= (batch_size+1):
                        callback(records, remaining=total_remaining, line=line_number, result_id=result_id)
                        total_remaining -= (len(records)-1)
                        records = [col_names]
                callback(records, remaining=total_remaining, line=line_number, result_id=result_id)

            tf.close()

        return True

    def get_batch_result_iter(self, job_id, batch_id, parse_csv=False, logger=None):
        """Return a line interator over the contents of a batch result document. If csv=True then parses the first line
        as the csv header and the iterator returns dicts."""
        status = self.batch_status(job_id, batch_id)
        if status['state'] != 'Completed':
            return None
        elif logger:
            if 'numberRecordsProcessed' in status:
                logger("Bulk batch %d processed %s records" % (batch_id, status['numberRecordsProcessed']))
            if 'numberRecordsFailed' in status:
                failed = int(status['numberRecordsFailed'])
                if failed > 0:
                    logger("Bulk batch %d had %d failed records" % (batch_id, failed))

        uri = self.endpoint + "/services/async/29.0/job/%s/batch/%s/result" % (job_id, batch_id)
        r = requests.get(uri, headers=self.headers(), stream=True)
        if parse_csv:
            return csv.DictReader(r.iter_lines(chunk_size=2048), delimiter=",", quotechar='"')
        else:
            return r.iter_lines(chunk_size=2048)


    def get_upload_results(self, job_id, batch_id, callback = dump_results, batch_size=0, logger=None):
        job_id = job_id or self.lookup_job_id(batch_id)

        if not self.is_batch_done(job_id, batch_id):
            return False
        http = Http()
        uri = self.endpoint + "/services/async/29.0/job/%s/batch/%s/result" % (job_id, batch_id)
        resp, content = http.request(uri, method="GET", headers=self.headers())

        tf = TemporaryFile()
        tf.write(content)

        total_remaining = self.count_file_lines(tf)
        if logger:
            logger("Total records: %d" % total_remaining)
        tf.seek(0)

        records = []
        line_number = 0
        col_names = []
        reader = csv.reader(tf, delimiter=",", quotechar='"')
        for row in reader:
            line_number += 1
            records.append(UploadResult(*row))
            if len(records) == 1:
                col_names = records[0]
            if batch_size > 0 and len(records) >= (batch_size+1):
                callback(records, total_remaining, line_number)
                total_remaining -= (len(records)-1)
                records = [col_names]
        callback(records, total_remaining, line_number)

        tf.close()

        return True

    def parse_csv(self, tf, callback, batch_size, total_remaining):
        print batch_size, total_remaining
        records = []
        line_number = 0
        col_names = []
        reader = csv.reader(tf, delimiter=",", quotechar='"')
        for row in reader:
            line_number += 1
            records.append(row)
            if len(records) == 1:
                col_names = records[0]
            if batch_size > 0 and len(records) >= (batch_size+1):
                callback(records, total_remaining, line_number)
                total_remaining -= (len(records)-1)
                records = [col_names]
        return records, total_remaining

    def count_file_lines(self, tf):
        tf.seek(0)
        buffer = bytearray(2048)
        lines = 0

        quotes = 0
        while tf.readinto(buffer) > 0:
            quoteChar = ord('"')
            newline = ord('\n')
            for c in buffer:
                if c == quoteChar:
                    quotes += 1
                elif c == newline:
                    if (quotes % 2) == 0:
                        lines += 1
                        quotes = 0

        return lines
        


import csv
from cStringIO import StringIO

class CsvDictsAdapter(object):
    """Provide a DataChange generator and it provides a file-like object which returns csv data"""
    def __init__(self, source_generator):
        self.source = source_generator
        self.buffer = StringIO()
        self.csv = None
        self.add_header = False

    def __iter__(self):
        return self

    def write_header(self):
        self.add_header = True

    def next(self):
        row = self.source.next()

        self.buffer.truncate(0)
        self.buffer.seek(0)

        if not self.csv:
            self.csv = csv.DictWriter(self.buffer, row.keys(), quoting = csv.QUOTE_NONNUMERIC)
            self.csv.writeheader()
        elif self.add_header:
            self.csv.writeheader()
            self.add_header = False

        self.csv.writerow(row)
        self.buffer.seek(0)
        return self.buffer.read()



import unittest

import salesforce_oauth_request

from salesforce_bulk import SalesforceBulk

class SalesforceBulkTest(unittest.TestCase):
    def __init__(self, testName, endpoint, sessionId):
        super(SalesforceBulkTest, self).__init__(testName)
        self.endpoint = endpoint
        self.sessionId = sessionId

    def setUp(self):
        self.jobs = []

    def tearDown(self):
        if hasattr(self, 'bulk'):
            for job_id in self.jobs:
                print "Closing job: %s" % job_id
                self.bulk.close_job(job_id)

    def test_raw_query(self):
        bulk = SalesforceBulk(self.endpoint, self.sessionId)
        self.bulk = bulk

        job_id = bulk.create_query_job("Contact")
        self.jobs.append(job_id)
        self.assertIsNotNone(re.match("\w+", job_id))

        batch_id = bulk.query(job_id, "Select Id,Name,Email from Contact Limit 1000")
        self.assertIsNotNone(re.match("\w+", batch_id))

        while not bulk.is_batch_done(job_id, batch_id):
            print "Job not done yet..."
            print bulk.batch_status(job_id, batch_id)
            time.sleep(2)

        self.results = ""
        def save_results(tfile, **kwargs):
            print "in save results"
            self.results = tfile.read()

        flag = bulk.get_batch_results(job_id, batch_id, callback = save_results)
        self.assertTrue(flag)
        self.assertTrue(len(self.results) > 0)
        self.assertIn('"', self.results)


    def test_csv_query(self):
        bulk = SalesforceBulk(self.endpoint, self.sessionId)
        self.bulk = bulk

        job_id = bulk.create_query_job("Account")
        self.jobs.append(job_id)
        self.assertIsNotNone(re.match("\w+", job_id))

        batch_id = bulk.query(job_id, "Select Id,Name,Description from Account Limit 10000")
        self.assertIsNotNone(re.match("\w+", batch_id))
        bulk.wait_for_batch(job_id, batch_id, timeout=120)

        self.results = None
        def save_results1(rows, **kwargs):
            self.results = rows

        flag = bulk.get_batch_results(job_id, batch_id, callback = save_results1, parse_csv=True)
        self.assertTrue(flag)
        results = self.results
        self.assertTrue(len(results) > 0)
        self.assertTrue(isinstance(results,list))
        self.assertEqual(results[0], ['Id','Name','Description'])
        self.assertTrue(len(results) > 3)

        self.results = None
        self.callback_count = 0
        def save_results2(rows, **kwargs):
            self.results = rows
            print rows
            self.callback_count += 1

        batch = len(results) / 3
        self.callback_count = 0
        flag = bulk.get_batch_results(job_id, batch_id, callback = save_results2, parse_csv=True, batch_size=batch)
        self.assertTrue(self.callback_count >= 3)


    def test_csv_upload(self):
        bulk = SalesforceBulk(self.endpoint, self.sessionId)
        self.bulk = bulk

        job_id = bulk.create_insert_job("Contact")
        self.jobs.append(job_id)
        self.assertIsNotNone(re.match("\w+", job_id))

        batch_ids = []
        content = open("example.csv").read()
        for i in range(5):
            batch_id = bulk.query(job_id, content)
            self.assertIsNotNone(re.match("\w+", batch_id))
            batch_ids.append(batch_id)

        for batch_id in batch_ids:
            bulk.wait_for_batch(job_id, batch_id, timeout=120)

        self.results = None
        def save_results1(rows, failed, remaining):
            self.results = rows

        for batch_id in batch_ids:
            flag = bulk.get_upload_results(job_id, batch_id, callback = save_results1)
            self.assertTrue(flag)
            results = self.results
            self.assertTrue(len(results) > 0)
            self.assertTrue(isinstance(results,list))
            self.assertEqual(results[0], UploadResult('Id','Success','Created','Error'))
            self.assertEqual(len(results), 3)

        self.results = None
        self.callback_count = 0
        def save_results2(rows, failed, remaining):
            self.results = rows
            self.callback_count += 1

        batch = len(results) / 3
        self.callback_count = 0
        flag = bulk.get_upload_results(job_id, batch_id, callback = save_results2, batch_size=batch)
        self.assertTrue(self.callback_count >= 3)


if __name__ == '__main__':
    username = raw_input("Salesforce username: ")
    password = raw_input("Salesforce password: ")

    login = salesforce_oauth_request.login(username=username, password=password)
    endpoint = login['endpoint']
    sessionId = login['access_token']

    suite = unittest.TestSuite()
    suite.addTest(SalesforceBulkTest("test_csv_upload", endpoint, sessionId))
    unittest.TextTestRunner().run(suite)








#!/usr/bin/env python

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

packages = [
    'salesforce_bulk',
]

requires = [
    'requests>=2.2.1',
]

with open('README.md') as f:
    readme = f.read()
with open('LICENSE') as f:
    license = f.read()

setup(
    name='salesforce-bulk',
    version='1.0.0',
    description='Python interface to the Salesforce.com Bulk API.',
    long_description=readme,
    author='Scott Persinger',
    author_email='scottp@heroku.com',
    url='https://github.com/heroku/salesforce-bulk',
    packages=packages,
    package_data={'': ['LICENSE']},
    include_package_data=True,
    install_requires=requires,
    license=license,
    zip_safe=False,
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ),
)
